///////////////////////////////////////////////////////////
//                     GGBuildTools                      //
//         Gipat Group       //
//             Copyright (C) 2007 Gipat Group            //
//                             //
//    Gipat Group's opened EI-editor-utility license     //
//                       1.1                       //
//                                                       //
//                  www.gipatgroup.org                   //
///////////////////////////////////////////////////////////

//      , ....  :
// 1) Immortal (ImmortalGAD@rambler.ru)
// 2) Sagrer (sagrer@yandex.ru)

////////////////////////////////////////////////////////////////////////.
//
//   Unit Archiver .       
//
////////////////////////////////////////////////////////////////////////.
unit Archiver;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, SconfFormat, ExtraFileUtilsLCL, process;

const
  ARCHIVER_NAME_7ZIP = '7z';
  ARCHIVER_7ZIP_EXTENSION = '.7z';
  ARCHIVER_7ZIP_DefUtilPath = '.\..\ExtTools\7zr.exe';

type

  //    ,    ,    
  //  
  TArchiver = class
  public
    UseSubfolder  :Boolean;        //   
    Name          :String;         //  
    SourceDir     :String;         //   
    OutputDir     :String;         //   
    ErrorMessage : AnsiString;     //    .
    UtilPath : AnsiString;    //   .
    //-...
    constructor Create; virtual;
    destructor Destroy; override;
    // .
    function GetFullArchiveName() : AnsiString; virtual; abstract;   //    .
    function Pack() : Boolean; virtual; abstract;    //   .

  end;
  
  //    ,   7zr.exe
  T7ZipArchiver = class(TArchiver)
  public
    //-...
    constructor Create; override;
    destructor Destroy; override;
    // .
    function GetFullArchiveName() : AnsiString; override;   //    .
    function Pack() : Boolean; override;   //      

  end;
  
  //  -.     .
  function GetArchiverByType(const typeA :String):TArchiver;
  
implementation

/////////////////////////////////////////////
//                                  //
/////////////////////////////////////////////

//  -.     .
function GetArchiverByType(const typeA :String):TArchiver;
begin
  //   ,   .
  Result := nil;
  if (ScF_ArchiveType_7z = typeA) then begin
    Result := T7ZipArchiver.Create;
  end;
end;

/////////////////////////////////////////////
//                TArchiver                //
/////////////////////////////////////////////

//-----------------------------------------//
//        -...      //
//-----------------------------------------//

constructor TArchiver.Create;
begin
  //  ...

  
  //    ...
  Self.UtilPath := '';
  Self.ErrorMessage := '';

  //   ...

end;

destructor TArchiver.Destroy;
begin
  // 

  //  
  inherited;
end;

/////////////////////////////////////////////
//              T7ZipArchiver              //
/////////////////////////////////////////////

//-----------------------------------------//
//        -...      //
//-----------------------------------------//

constructor T7ZipArchiver.Create;
begin
  //  .
  inherited;
  
  //  ...

  //    ...

  //   ...

end;

destructor T7ZipArchiver.Destroy;
begin
  // 

  //  
  inherited;
end;

//------------------------------------------//
//             ...            //
//------------------------------------------//

function T7ZipArchiver.GetFullArchiveName() : AnsiString;
//    .
begin
  Result := ResolveLocalPath(Name, OutputDir) + ARCHIVER_7ZIP_EXTENSION;
end;

//      
function T7ZipArchiver.Pack() : Boolean;
var
  path :String;        //   7zr.exe
  args :String;        // 
  packagePath :String; //    
  process :TProcess;   //    
  
begin
  //.
  Result := true;
  
  //    .
  if FileExists(Self.UtilPath) = false then begin
    //        -   .
    path := ResolveLocalPath(ARCHIVER_7ZIP_DefUtilPath, ExtractFilePath(ParamStr(0)));
    if FileExists(path) = false then begin
      // . .
      Result := false;
      Self.ErrorMessage := 'Can''t find 7zr utility! File '+path+' is not exists.';
    end;
  end
  else begin
    // ,  .
    path := Self.UtilPath;
  end;
  
  //  - .
  if Result = true then begin
    //    
    packagePath := Self.GetFullArchiveName();
    //   .
    args := Format('a -mx=9 -ms=on "%s" "%s"', [packagePath, SourceDir]);

    //  ,    .  .
    if FileExists(packagePath) then begin
      DeleteFile(packagePath);
    end;

    //   
    process := TProcess.Create(nil);
    process.CommandLine := Format('"%s" %s',[path, args]);
    //    .
    process.Options := [poNoConsole, poWaitOnExit];
    // 
    process.Execute;
    process.Free;

    //   .      , 
    //    .   
    if process.ExitStatus <> 0 then begin
      Result := false;
      Self.ErrorMessage := 'Extern Process execution faild! '+path+' was executed.';
    end;
  end;   //if Result = true
end;

end.

