///////////////////////////////////////////////////////////
//                     GGBuildTools                      //
//         Gipat Group       //
//             Copyright (C) 2007 Gipat Group            //
//                             //
//    Gipat Group's opened EI-editor-utility license     //
//                       1.0                       //
//                                                       //
//                  www.gipatgroup.org                   //
///////////////////////////////////////////////////////////

//      , ....  :
// 1) Sagrer (sagrer@yandex.ru)

////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////
//             Flist       //
//            .   //
//             //
//                  ...                   //
//                                                //
//                - 1                //
////////////////////////////////////////////////////

unit FlistFormat;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Windows, LCLAnIniFile, crc, ExtraFileUtilsLcl, ExtraFunctionsLcl;
  
const
  Flist_FormatVersion = 1;                //     .
  //   ....
  Flist_Def_PathsAreAbsolute = false;           //    .
  Flist_Def_AddSize = true;                     //      .
  Flist_Def_AddCRC32 = true;                    //      CRC32.
  Flist_Def_DoConsoleOutput = false;            //      .
  // 
  Flist_Filetype_File = 0;                //  - .
  Flist_Filetype_Directory = 1;           //  - .
  Flist_Filetype_Unknown = 2;             //  -  .
  // CheckFile
  Flist_CheckFile_Identical = 0;          // .
  Flist_CheckFile_MismatchedCRC32 = 1;    //  CRC32.
  Flist_CheckFile_MismatchedSize = 2;     //  .
  Flist_CheckFile_ElementNotExists = 3;   //     .
  Flist_CheckFile_FileNotExists = 4;      //    .
  Flist_CheckFile_NothingExists = 5;      //           .
  Flist_CheckFile_Error = 6;              //. ( - ).
  Flist_CheckFile_NonIdentical = 7;       //  ( -    ).


type

  RFlistFile = record                     //      .
    FilePath : AnsiString;                //  .
    FileType : byte;                      //  - Flist_Filetype_* .
    FileSize : Int64;                   // .
    CRC32 : Cardinal;                     //CRC32 .
  end;
  
  RComparedElement = record               //     .
    FilePath : AnsiString;                //  .
    FileType : byte;                      //  - Flist_Filetype_* .
    CheckResult : byte;                   //     \.
  end;
  
  AComparedList = array of RComparedElement;   //     .

  TFlistFormat = class                    //      Flist (*.flist).
  private
    // .
    function ParseFileType(const InputStr : AnsiString) : byte;       //       .
    function ReturnFileTypeStr(const InputCode : byte) : AnsiString;  //      .
  public
    //
    //[Main]
    PathsAreAbsolute : Boolean;           //   .  false -         *.flist
    AddSize : Boolean;                    //     .
    AddCRC32 : Boolean;                   //  crc32 .
    FilesNum : LongInt;                   //   .
    //[Files]
    FileList : array of RFlistFile;       //    (  ).
    //
    ErrorMessage : AnsiString;            //    
    FileLoaded : boolean;                 // true -    - .
    LoadedFileName : AnsiString;          //     .
    GlobalRoot : AnsiString;              //   -       flist.
    DoConsoleOutput : boolean;            //  true  -       .
    
    //-...
    constructor Create;
    destructor Destroy; override;
    
    // ...
    procedure Clear();                                              //  .
    function GetGlobalRoot() : AnsiString;                          // .
    function Load(const FileName : AnsiString) : boolean;           //   .
    function Save(const FileName : AnsiString) : boolean;           //     .
    function ElementNameExists(const ElementName : AnsiString) : boolean;   //     .
    function GetElementNumber(const ElementName : AnsiString) : Integer;   //      .
    function FileExistsHere(const FileName : AnsiString) : boolean;     //     .
    function GetFileNumberHere(const FileName : AnsiString) : Integer;   //      .
    function DirExistsHere(const DirName : AnsiString) : boolean;       //     .
    function GetDirNumberHere(const DirName : AnsiString) : Integer;   //     .
    function GetRealElementName(const ElemName : AnsiString) : AnsiString;   //    
    function GetRealElementName(const ElemNum : Integer) : AnsiString;   //      -   .
    function AddFile(const FileName : AnsiString) : boolean;        // .
    function AddRealFile(const FileName : AnsiString) : boolean;    //      (   ).
    function AddDirectory(const DirName : AnsiString) : boolean;    //  .
    function AddRealDirectory(const DirName : AnsiString; const Recursive : boolean) : boolean;  //        .
    function RemoveFile(const FileName : AnsiString) : boolean;     //   .
    function RemoveDirectory(const DirName : AnsiString) : boolean;  //  (     ).
    function CheckFile(const ListFileName, RealFileName : AnsiString) : byte;   //       .      Flist_CheckFile_*.   .
    procedure CompareDirs(const ListDirName, RealDirName : AnsiString; var ComparedList : AComparedList);   // 2 ,    ComparedList.
    function CompareWithFlist(var OtherFlist : TFlistFormat; const ListDirName : AnsiString; var ComparedList : AComparedList) : boolean;   //    ,   (  )        .
    function CompareWithFlist(const OtherFlistFile, ListDirName : AnsiString; var ComparedList : AComparedList) : boolean;   //    ,   (  )        .
    function ConvertToSimpleComparedList(var ComparedList : AComparedList) : Boolean;  //      -         ,      .

  end;
  
//  ,   %).
function Flist_CheckFile_2Text(const Code : Byte) : AnsiString;    //     Flist_CheckFile_*

implementation

/////////////////////////////////////////////
//              TFlistFormat               //
/////////////////////////////////////////////

//-----------------------------------------//
//        -...      //
//-----------------------------------------//

constructor TFlistFormat.Create;
begin
  //  ...
  Self.Clear();
  ErrorMessage := 'none';
  DoConsoleOutput := Flist_Def_DoConsoleOutput;

  //   ...

end;

destructor TFlistFormat.Destroy;
begin
  // 
  SetLength(FileList,0);

  //  
  inherited;
end;

//------------------------------------------//
//            ...             //
//------------------------------------------//

function TFlistFormat.ParseFileType(const InputStr : AnsiString) : byte;
//       .
begin
  if LowerCase(InputStr) = 'file' then begin
    //
    Result := Flist_Filetype_File;
  end
  else if LowerCase(InputStr) = 'dir' then begin
    //
    Result := Flist_Filetype_Directory;
  end
  else begin
    //-
    Result := Flist_Filetype_Unknown;
  end;
end;

function TFlistFormat.ReturnFileTypeStr(const InputCode : byte) : AnsiString;
//      .
begin
  if InputCode = Flist_Filetype_File then begin
    //
    Result := 'File';
  end
  else if InputCode = Flist_Filetype_Directory then begin
    //
    Result := 'Dir';
  end
  else begin
    //-
    Result := 'Unknown';
  end;
end;

//------------------------------------------//
//            ...             //
//------------------------------------------//

procedure TFlistFormat.Clear();
//  .
begin
  // -   ...
  PathsAreAbsolute := Flist_Def_PathsAreAbsolute;
  AddSize := Flist_Def_AddSize;
  AddCRC32 := Flist_Def_AddCRC32;
  FilesNum := 0;
  //    0 
  SetLength(FileList,0);
  //  .
  FileLoaded := false;
  LoadedFileName := '';
  GlobalRoot := '';
end;

function TFlistFormat.GetGlobalRoot() : AnsiString;
// .
begin
  if Self.GlobalRoot = '' then begin
    //    .
    if Self.FileLoaded = true then begin
      //  -  -     .
      Result := ExtractFilePath(Self.LoadedFileName);
    end
    else begin
      //     .
      Result := GetCurrentDir();
    end;
  end
  else begin
    //      -   .
    Result := Self.GlobalRoot;
  end;
end;

function TFlistFormat.Load(const FileName : AnsiString) : boolean;
//    .
var
  AnIniFile : TAnIniFile;
  AllOk : boolean;
  IniValueList : AIniValueList;
  I, J : Integer;
  
begin
  //.
  AnIniFile := TAnIniFile.Create;
  SetLength(IniValueList,0,0);  //       [Files]
  Result := false;
  if Self.DoConsoleOutput = true then begin
    Writeln('Loading file '+FileName+' ... ');
  end;
  
  //  ...
  Self.Clear();
  
  // ...
  AllOk := FileExists(FileName);
  if AllOk = false then begin
    //   ...
    ErrorMessage := 'File '+FileName+' is not exists.';
  end
  else begin
    //    ... ...
    AllOk := AnIniFile.Load(FileName);
    if AllOk = false then begin
      //   .
      ErrorMessage := 'Error, loading file '+FileName;
    end;
  end;
  
  if AllOk = true then begin
    //, .  ...
    if AnIniFile.ReadString('Main', 'FormatName') <> 'FileList' then begin
      //  FileList-.
      AllOk := false;
      ErrorMessage := 'File '+FileName+' is not FileList file.';
    end;
  end;

  if AllOk = true then begin
    //  ...

    if AnIniFile.ReadInteger('Main', 'FormatVersion') > Flist_FormatVersion then begin
      //       .
      AllOk := false;
      ErrorMessage := 'File '+FileName+' has more new version ('+AnIniFile.ReadString('Main', 'FormatVersion')+') then i know ('+IntToStr(Flist_FormatVersion)+'). Aborted loading. Try update to a new version of this software.';
    end;
  end;
  
  if AllOk = true then begin
    //  ...
    //      -     .

    //[Main]
    PathsAreAbsolute := AnIniFile.ReadBool('Main','PathsAreAbsolute');
    AddSize := AnIniFile.ReadBool('Main','AddSize');
    AddCRC32 := AnIniFile.ReadBool('Main','AddCRC32');
    //FilesNum := AnIniFile.ReadInteger('Main','FilesNum');      //,     .

    //[Files]
    AnIniFile.ReadIniValueList('Files',IniValueList);   // ...
    //    ...
    FilesNum := Length(IniValueList);
    //   ...
    SetLength(FileList,FilesNum);
    //  ...
    if FilesNum > 0 then begin
      //    .
      for I := 0 to FilesNum-1 do begin
        for J := 0 to Length(IniValueList[I])-1 do begin
          // .
          if IniValueList[I,J].Name = 'File' then begin
            FileList[I].FilePath := IniValueList[I,J].Value;
          end
          else if IniValueList[I,J].Name = 'Type' then begin
            FileList[I].FileType := ParseFileType(IniValueList[I,J].Value);
          end
          else if IniValueList[I,J].Name = 'Size' then begin
            FileList[I].FileSize := StrToInt(IniValueList[I,J].Value);
          end
          else if IniValueList[I,J].Name = 'CRC32' then begin
            FileList[I].CRC32 := StrToInt('$'+IniValueList[I,J].Value);     //  - ..     HEX.
          end;
        end;
      end;
    end;

    //,      .       .
    Result := true;
    if Self.DoConsoleOutput = true then begin
      Writeln('Loaded file '+FileName);
    end;
    FileLoaded := true;
    LoadedFileName := FileName;
  end;
  
  // .
  AnIniFile.Free;
  SetLength(IniValueList,0,0);
end;

function TFlistFormat.Save(const FileName : AnsiString) : boolean;
//     .
var
  AnIniFile : TAnIniFile;
  I, J, ValElemNum : Integer;
  IniValueList : AIniValueList;
  
begin
  //.
  AnIniFile := TAnIniFile.Create;
  SetLength(IniValueList,0,0);
  Result := false;
  if Self.DoConsoleOutput = true then begin
    Writeln('Saving file '+FileName+' ... ');
  end;
  
  //  ...
  AnIniFile.MakNewFile;
  // ...
  //[Main]
  AnIniFile.WriteString('Main', 'FormatName', 'FileList');
  AnIniFile.WriteInteger('Main', 'FormatVersion', Flist_FormatVersion);
  AnIniFile.WriteBool('Main', 'PathsAreAbsolute', PathsAreAbsolute);
  AnIniFile.WriteBool('Main', 'AddSize', AddSize);
  AnIniFile.WriteBool('Main', 'AddCRC32', AddCRC32);
  AnIniFile.WriteInteger('Main', 'FilesNum', FilesNum);
  
  //[Files]
  //      "" IniValueList
  ValElemNum := 2;      //.
  if AddSize = true then ValElemNum := ValElemNum+1;
  if AddCRC32 = true then ValElemNum := ValElemNum+1;
  //    IniValueList
  SetLength(IniValueList,FilesNum,ValElemNum);
  //      IniValueList
  for I := 0 to FilesNum-1 do begin
    J := 0;
    //File
    IniValueList[I,J].Name := 'File';
    IniValueList[I,J].Value := FileList[I].FilePath;
    J := J+1;
    //Type
    IniValueList[I,J].Name := 'Type';
    IniValueList[I,J].Value := ReturnFileTypeStr(FileList[I].FileType);
    J := J+1;
    //Size
    if AddSize = true then begin
      IniValueList[I,J].Name := 'Size';
      IniValueList[I,J].Value := IntToStr(FileList[I].FileSize);
      J := J+1;
    end;
    //CRC32
    if AddCRC32 = true then begin
      IniValueList[I,J].Name := 'CRC32';
      IniValueList[I,J].Value := IntToHex(FileList[I].CRC32,8);
    end;
  end;
  //    ...
  AnIniFile.WriteIniValueList('Files',IniValueList);
  
  //   ...
  Result := AnIniFile.Save(FileName);
  //   -     .
  if Result = true then begin
    FileLoaded := true;
    LoadedFileName := FileName;
    if Self.DoConsoleOutput = true then begin
      Writeln('Saved file '+FileName);
    end;
  end;
  
  //  ( ).
  AnIniFile.Free;
  SetLength(IniValueList,0,0);
end;

function TFlistFormat.ElementNameExists(const ElementName : AnsiString) : boolean;
//     .
var
  I : Integer;
  TempElemName : AnsiString;
  
begin
  //.
  Result := false;
  I := -1;
  TempElemName := UpperCase(DelLastSlash(ElementName));
  
  //              .
  repeat
  
    I := I+1;
    //,      -  ,   - .
    if UpperCase(DelLastSlash(FileList[I].FilePath)) = TempElemName then begin
      Result := true;     //, ,        true.
    end;
    
  until (Result = true) or (I = FilesNum-1);
  
end;

function TFlistFormat.GetElementNumber(const ElementName : AnsiString) : Integer;
//      .
var
  I : Integer;
  TempElemName : AnsiString;

begin
  //.
  Result := -1;
  I := -1;
  TempElemName := UpperCase(DelLastSlash(ElementName));

  //              .
  repeat

    I := I+1;
    //,      -  ,   - .
    if UpperCase(DelLastSlash(FileList[I].FilePath)) = TempElemName then begin
      Result := I;     //, ,        true.
    end;

  until (Result = I) or (I = FilesNum-1);

end;

function TFlistFormat.FileExistsHere(const FileName : AnsiString) : boolean;
//     .
var
  I : Integer;
  TempFileName : AnsiString;
  
begin
  //.
  Result := false;
  I := -1;
  TempFileName := UpperCase(FileName);

  //              .
  repeat

    I := I+1;
    
    if FileList[I].FileType = Flist_Filetype_File then begin
      //   .
      if UpperCase(FileList[I].FilePath) = TempFileName then begin
        Result := true;     //, ,        true.
      end;
    end;

  until (Result = true) or (I = FilesNum-1);

end;

function TFlistFormat.GetFileNumberHere(const FileName : AnsiString) : Integer;
//      .
var
  I : Integer;
  TempFileName : AnsiString;

begin
  //.
  Result := -1;
  I := -1;
  TempFileName := UpperCase(FileName);

  //              .
  repeat

    I := I+1;

    if FileList[I].FileType = Flist_Filetype_File then begin
      //   .
      if UpperCase(FileList[I].FilePath) = TempFileName then begin
        Result := I;     //, ,        true.
      end;
    end;

  until (Result = I) or (I = FilesNum-1);

end;

function TFlistFormat.DirExistsHere(const DirName : AnsiString) : boolean;
//     .
var
  I : Integer;
  TempDirName : AnsiString;

begin
  //.
  Result := false;
  I := -1;
  TempDirName := DelLastSlash(UpperCase(DirName));

  //              .
  repeat

    I := I+1;

    if FileList[I].FileType = Flist_Filetype_Directory then begin
      //   .
      if UpperCase(DelLastSlash(FileList[I].FilePath)) = TempDirName then begin
        Result := true;     //, ,        true.
      end;
    end;

  until (Result = true) or (I = FilesNum-1);

end;

function TFlistFormat.GetDirNumberHere(const DirName : AnsiString) : Integer;
//     .
var
  I : Integer;
  TempDirName : AnsiString;

begin
  //.
  Result := -1;
  I := -1;
  TempDirName := DelLastSlash(UpperCase(DirName));

  //              .
  repeat

    I := I+1;

    if FileList[I].FileType = Flist_Filetype_Directory then begin
      //   .
      if UpperCase(DelLastSlash(FileList[I].FilePath)) = TempDirName then begin
        Result := I;     //, ,        true.
      end;
    end;

  until (Result = I) or (I = FilesNum-1);

end;

function TFlistFormat.GetRealElementName(const ElemName : AnsiString) : AnsiString;
//    
begin
  //  ....
  if IsLocalPath(ElemName) = true then begin
    //   - .
    Result := ResolveLocalPath(ElemName,Self.GetGlobalRoot());
  end
  else begin
    //  .
    Result := ElemName;
  end;
end;

function TFlistFormat.GetRealElementName(const ElemNum : Integer) : AnsiString;
//      -   .
begin
  //      ...
  Result := GetRealElementName(Self.FileList[ElemNum].FilePath);
end;

function TFlistFormat.AddFile(const FileName : AnsiString) : boolean;
// .
var
  CurrFileNum : Integer;
  LocalName, GlobalName : AnsiString;
  
begin
  //.
  Result := false;
  CurrFileNum := FilesNum;   //          .
  
  //    .
  FilesNum := FilesNum+1;
  SetLength(FileList,FilesNum);
  
  //  ....
  if IsLocalPath(FileName) = true then begin
    //   - .
    GlobalName := ResolveLocalPath(FileName,Self.GetGlobalRoot());
  end
  else begin
    //  .
    GlobalName := FileName;
  end;

  //  -   .
  if Self.PathsAreAbsolute = false then begin
    LocalName := GetLocalPath(Self.GetGlobalRoot(),GlobalName);
  end;
  
  //   .
  //FilePath
  if Self.PathsAreAbsolute = true then begin
    //   .
    FileList[CurrFileNum].FilePath := GlobalName;
  end
  else begin
    //   
    FileList[CurrFileNum].FilePath := LocalName;
  end;
  //FileType
  FileList[CurrFileNum].FileType := Flist_Filetype_File;
  //FileSize
  if Self.AddSize = true then begin
    FileList[CurrFileNum].FileSize := 0;
  end;
  //CRC32
  if Self.AddCRC32 = true then begin
    FileList[CurrFileNum].CRC32 := crc32(0, nil, 0);         // ""  CRC32.
  end;
  
  // .
  Result := true;
end;

function TFlistFormat.AddRealFile(const FileName : AnsiString) : boolean;
//      (   ).
var
  CurrFileNum : Integer;
  LocalName, GlobalName : AnsiString;
  
begin
  //    TFlistFormat.AddFile     .
  //.
  Result := false;
  CurrFileNum := FilesNum;   //          .
  
  //  ....
  if IsLocalPath(FileName) = true then begin
    //   - .
    GlobalName := ResolveLocalPath(FileName,Self.GetGlobalRoot());
  end
  else begin
    //  .
    GlobalName := FileName;
  end;
  
  //  -   .
  if Self.PathsAreAbsolute = false then begin
    LocalName := GetLocalPath(Self.GetGlobalRoot(),GlobalName);
  end;

  //    .
  FilesNum := FilesNum+1;
  SetLength(FileList,FilesNum);

  //   .
  //FilePath
  if Self.PathsAreAbsolute = true then begin
    //   .
    FileList[CurrFileNum].FilePath := GlobalName;
  end
  else begin
    //   
    FileList[CurrFileNum].FilePath := LocalName;
  end;
  //FileType
  FileList[CurrFileNum].FileType := Flist_Filetype_File;
  //FileSize
  if Self.AddSize = true then begin
    FileList[CurrFileNum].FileSize := GetFileSize(GlobalName);       // .    -  .
  end;
  //CRC32
  if Self.AddCRC32 = true then begin
    FileList[CurrFileNum].CRC32 := FileToCRC32(GlobalName,DoConsoleOutput);    // CRC32  .    -  "" .
  end;

  // .
  Result := true;
end;

function TFlistFormat.AddDirectory(const DirName : AnsiString) : boolean;
//  .
var
  CurrFileNum : Integer;
  LocalName, GlobalName : AnsiString;

begin
  //.
  Result := false;
  CurrFileNum := FilesNum;   //          .

  //    .
  FilesNum := FilesNum+1;
  SetLength(FileList,FilesNum);
  
  //  ....
  if IsLocalPath(DirName) = true then begin
    //   - .
    GlobalName := DelLastSlash(ResolveLocalPath(DirName,Self.GetGlobalRoot()));
  end
  else begin
    //  .
    GlobalName := DelLastSlash(DirName);
  end;

  //  -   .
  if Self.PathsAreAbsolute = false then begin
    LocalName := GetLocalPath(Self.GetGlobalRoot(),GlobalName);
  end;

  //   .
  //FilePath
  if Self.PathsAreAbsolute = true then begin
    //   .
    FileList[CurrFileNum].FilePath := GlobalName;
  end
  else begin
    //   
    FileList[CurrFileNum].FilePath := LocalName;
  end;
  //FileType
  FileList[CurrFileNum].FileType := Flist_Filetype_Directory;

  // .
  Result := true;
end;

function TFlistFormat.AddRealDirectory(const DirName : AnsiString; const Recursive : boolean) : boolean;
//        .
var
  AddingFiles : LongInt;
  LocalName, GlobalName : AnsiString;
  
  function ReallyAddDirectory(const DirName : AnsiString; const Recursive : boolean) : boolean;
  //,     ...
  var
    SearchRec : TSearchRec;
    LocalDirName : AnsiString;
  begin
    //.
    Result := false;
    
    //  -   .
    if Self.PathsAreAbsolute = false then begin
      LocalDirName := GetLocalPath(Self.GetGlobalRoot(),DirName);
    end;
    
    //  ...
    if FindFirst(AddLastSlash(DirName)+'*', faAnyFile and faDirectory, SearchRec) = 0 then begin
      //   .
      repeat
        //         %).
        if (SearchRec.Name <> '.') and (SearchRec.Name <> '..') then begin
          //- .
          if (SearchRec.Attr and faDirectory) <> faDirectory then begin
            //  .        -    %).  
            //  "  "  -  -    %).
            //,    - . .

            //   .
            //FilePath
            if Self.PathsAreAbsolute = true then begin
              //   .
              FileList[FilesNum].FilePath := AddLastSlash(DirName)+SearchRec.Name;
            end
            else begin
              //   
              FileList[FilesNum].FilePath := AddLastSlash(LocalDirName)+SearchRec.Name;
            end;
            //FileType
            FileList[FilesNum].FileType := Flist_Filetype_File;
            //FileSize
            if Self.AddSize = true then begin
              FileList[FilesNum].FileSize := GetFileSize(AddLastSlash(DirName)+SearchRec.Name);       // .    -  .
            end;
            //CRC32
            if Self.AddCRC32 = true then begin
              FileList[FilesNum].CRC32 := FileToCRC32(AddLastSlash(DirName)+SearchRec.Name,DoConsoleOutput);    // CRC32  .    -  "" .
            end;
            
            //+1      .
            FilesNum := FilesNum+1;
            
          end
          else begin
            //    - .  - ...

            //   .
            //FilePath
            if Self.PathsAreAbsolute = true then begin
              //   .
              FileList[FilesNum].FilePath := AddLastSlash(DirName)+SearchRec.Name;
            end
            else begin
              //   
              FileList[FilesNum].FilePath := AddLastSlash(LocalDirName)+SearchRec.Name;
            end;
            //FileType
            FileList[FilesNum].FileType := Flist_Filetype_Directory;
            
            //+1      .
            FilesNum := FilesNum+1;

            //   -   .
            if Recursive = true then begin
              ReallyAddDirectory(AddLastSlash(DirName)+SearchRec.Name, Recursive);
            end;
          end;
        end;
        
      until FindNext(SearchRec) <> 0;
      
      // .
      SysUtils.FindClose(SearchRec);
      // .
      Result := true;
    end;
  end;
  
begin
  //   "".           
  //  -    "" .    -  
  //   (   ,   ),   
  // ,      ,      
  // .       ,   
  //        .     
  //         -     .
  
  //.
  Result := false;
  
  //  ....
  if IsLocalPath(DirName) = true then begin
    //   - .
    GlobalName := DelLastSlash(ResolveLocalPath(DirName,Self.GetGlobalRoot()));
  end
  else begin
    //  .
    GlobalName := DelLastSlash(DirName);
  end;

  //  -   .
  if Self.PathsAreAbsolute = false then begin
    LocalName := GetLocalPath(Self.GetGlobalRoot(),GlobalName);
  end;
  
  //, -    ...
  AddingFiles := GetFilesNumber(GlobalName,Recursive)+1;       //+1 -     DirName
  //    ... FilesNum   -     .
  SetLength(FileList,FilesNum+AddingFiles);

  //  .
  //FilePath
  if Self.PathsAreAbsolute = true then begin
    //   .
    FileList[FilesNum].FilePath := GlobalName;
  end
  else begin
    //   
    FileList[FilesNum].FilePath := LocalName;
  end;
  //FileType
  FileList[FilesNum].FileType := Flist_Filetype_Directory;
  //+1    .
  FilesNum := FilesNum+1;
  
  //.        ,       .
  Result := ReallyAddDirectory(GlobalName, Recursive);
  
end;

function TFlistFormat.RemoveFile(const FileName : AnsiString) : boolean;
//   .
var
  RealSearchPath, TempStr : AnsiString;
  Finded : boolean;
  I, J : Integer;
  NewFileList : array of RFlistFile;
  DeletingElem, DummyElem : RFlistFile;
  
begin
  //
  Result := false;
  
  //                  .
  if Self.PathsAreAbsolute = true then begin
    //   .
    if IsLocalPath(FileName) = true then begin
      //   - .
      RealSearchPath := ResolveLocalPath(FileName,Self.GetGlobalRoot());
    end
    else begin
      //  .
      RealSearchPath := FileName;
    end;
  end
  else begin
    //   .
    if IsLocalPath(FileName) = true then begin
      //  .
      RealSearchPath := FileName;
    end
    else begin
      // 
      RealSearchPath := GetLocalPath(Self.GetGlobalRoot(),FileName);
    end;
  end;

  //    \      .
  Finded := false;
  I := -1;
  repeat
    I := I+1;
    if UpperCase(Self.FileList[I].FilePath) = UpperCase(RealSearchPath) then begin
      //.
      Finded := true;
    end;
  until (Finded = true) or (I = FilesNum-1);
  
  if Finded = true then begin
    //   ,      -    .
    if I <> (FilesNum-1) then begin
      Move(Self.FileList[I],DeletingElem,SizeOf(RFlistFile));    //    .
      Move(Self.FileList[I+1],Self.FileList[I],SizeOf(RFlistFile)*(FilesNum-(I+1)));    //   .
      Move(DeletingElem,Self.FileList[FilesNum-1],SizeOf(RFlistFile));     //             (   ).
      Move(DummyElem,DeletingElem,SizeOf(RFlistFile));     //       -                  (..     )  .
    end;
    //   .
    Self.FileList[FilesNum-1].CRC32 := 0;
    Self.FileList[FilesNum-1].FilePath := '';
    Self.FileList[FilesNum-1].FileSize := 0;
    Self.FileList[FilesNum-1].FileType := 0;
    //  .
    FilesNum := FilesNum-1;
    SetLength(Self.FileList,FilesNum);
    //Done.
    Result := true;
  end;
end;

function TFlistFormat.RemoveDirectory(const DirName : AnsiString) : boolean;
//  (     ).
var
  I : Integer;
  RealSearchPath : AnsiString;
  
begin
  //  :   ,   child- ( ..    
  //    -      ) -    , 
  //    .
  
  //.
  Result := false;
  
  //                  .
  if Self.PathsAreAbsolute = true then begin
    //   .
    if IsLocalPath(DirName) = true then begin
      //   - .
      RealSearchPath := ResolveLocalPath(DirName,Self.GetGlobalRoot());
    end
    else begin
      //  .
      RealSearchPath := DirName;
    end;
  end
  else begin
    //   .
    if IsLocalPath(DirName) = true then begin
      //  .
      RealSearchPath := DirName;
    end
    else begin
      // 
      RealSearchPath := GetLocalPath(Self.GetGlobalRoot(),DirName);
    end;
  end;
  
  //        ...
  I := -1;    //  .
  repeat
    I := I+1;   //.
    if I < Self.FilesNum then begin
      //     ,   ,      
      //       FilesNum...
      if IsDirsChild(RealSearchPath, Self.FileList[I].FilePath) = true then begin
        //    . .
        RemoveFile(Self.FileList[I].FilePath);
        //   .
        I := I-1;
      end;
    end;
  until I = Self.FilesNum-1;

end;

function TFlistFormat.CheckFile(const ListFileName, RealFileName : AnsiString) : byte;
//       .
//     Flist_CheckFile_*.   .
var
  IsDirectory : boolean;
  CurElemNum : Integer;
  
begin
  //.
  Result := Flist_CheckFile_Error;     //    ...
  
  //       .   .
  //,     ...
  if FileExists(RealFileName) = true then begin
    //    - .
    IsDirectory := false;
  end
  else if DirectoryExists(DelLastSlash(RealFileName)) = true then begin
    //    - .
    IsDirectory := true;
  end
  else begin
    //   ,  ,  .  -     .
    if ElementNameExists(RealFileName) = true then begin
      //   ,     .     
      // -    -  .
      Result := Flist_CheckFile_FileNotExists;
    end
    else begin
      //     ,    - .. .
      Result := Flist_CheckFile_NothingExists;
    end;
  end;
  
  if Result = Flist_CheckFile_Error then begin
    // Flist_CheckFile_Error            % )).
    //      - ,       .
    //    -     -     ...
    
    if IsDirectory = false then begin
      // .
      CurElemNum := GetFileNumberHere(ListFileName);
      if CurElemNum >= 0 then begin
        //   .
        // ,   ...
        if Self.AddSize = true then begin
          if GetFileSize(RealFileName) <> Self.FileList[CurElemNum].FileSize then begin
            // . .
            Result := Flist_CheckFile_MismatchedSize;
          end;
        end;
        
        //    -  ...
        if Result = Flist_CheckFile_Error then begin
          //  -  CRC32.
          if Self.AddCRC32 = true then begin
            if FileToCRC32(RealFileName) <> Self.FileList[CurElemNum].CRC32 then begin
              // . .
              Result := Flist_CheckFile_MismatchedCRC32;
            end;
          end;
        end;
        
        //       -   .
        if Result = Flist_CheckFile_Error then begin
          Result := Flist_CheckFile_Identical;
        end;
        
      end
      else begin
        //  ,    .
        Result := Flist_CheckFile_ElementNotExists;
      end;
    end;
    
  end
  else begin
    // .
    CurElemNum := GetDirNumberHere(ListFileName);
    if CurElemNum >= 0 then begin
      //   .         .        CRC32.
      Result := Flist_CheckFile_ElementNotExists;
    end;
  end;
end;

procedure TFlistFormat.CompareDirs(const ListDirName, RealDirName : AnsiString; var ComparedList : AComparedList);
// 2 ,    ComparedList.
var
  ListRealSize, I : Integer;
  AbsoluteName, ListName : AnsiString;
  MismatchFound : Boolean;
  
  procedure ReserveComparedListMem();
  //  - ""     ComparedList
  begin
    if ListRealSize >= Length(ComparedList) then begin
      //     .
      SetLength(ComparedList, Length(ComparedList)+50);
    end;
  end;
  
  function ElementReported(const ElemRepName : AnsiString) : Boolean;
  // -         .
  var
    I : Integer;
  begin
    Result := false;
    //...
    for I := 0 to ListRealSize-1 do begin
      if UpperCase(ComparedList[I].FilePath) = UpperCase(ElemRepName) then begin
        //.
        Result := true;
        Exit;
      end;
    end;
  end;
  
  procedure RecursiveCompareDir(const ListName, RealName : AnsiString);
  //        Flist.
  var
    SearchRec : TSearchRec;
    SearchElemNum : Integer;
    TmpCheckResult : Byte;
    
  begin
    //  -     . ,        ...
    if ElementReported(ListName) = false then begin
      //    .
      if DirExistsHere(ListName) = false then begin
        // -     .
        ReserveComparedListMem();
        ComparedList[ListRealSize].FilePath := FileList[I].FilePath;
        ComparedList[ListRealSize].FileType := Flist_Filetype_Directory;
        ComparedList[ListRealSize].CheckResult := Flist_CheckFile_ElementNotExists;
        ListRealSize := ListRealSize+1;
      end;
    end;

    if FindFirst(AddLastSlash(RealName)+'*', faAnyFile and faDirectory, SearchRec) = 0 then begin
      //    -    .
      repeat
        if (SearchRec.Name <> '.') and (SearchRec.Name <> '..') then begin
          //- .    .
          if (SearchRec.Attr and faDirectory) = faDirectory then begin
            // .   -    ..
            RecursiveCompareDir(AddLastSlash(ListName)+SearchRec.Name,AddLastSlash(RealName)+SearchRec.Name);
          end
          else begin
            // . ,      ...
            if ElementReported(AddLastSlash(ListName)+SearchRec.Name) = false then begin
              //     .      % ))).
              TmpCheckResult := Self.CheckFile(AddLastSlash(ListName)+SearchRec.Name,AddLastSlash(RealName)+SearchRec.Name);
              if TmpCheckResult <> Flist_CheckFile_Identical then begin
                //     ,     .
                ReserveComparedListMem();
                ComparedList[ListRealSize].FilePath := AddLastSlash(ListName)+SearchRec.Name;
                ComparedList[ListRealSize].FileType := Flist_Filetype_File;
                ComparedList[ListRealSize].CheckResult := TmpCheckResult;
                ListRealSize := ListRealSize+1;
              end;
            end;
          end;
        end;
      until FindNext(SearchRec) <> 0;
      
      //   -  .
      SysUtils.FindClose(SearchRec);
    end;
  end;
  
begin
  //     ,      ,     .
  //     -     - ,     
  // -   .
  
  // ComparedList.
  SetLength(ComparedList,0);
  
  //   ComparedList.
  ListRealSize := 0;     //      .
  SetLength(ComparedList,Length(Self.FileList)+50);   //    ,         .
  
  //    ...
  for I := 0 to FilesNum-1 do begin
    //  .
    MismatchFound := false;
    // -      ...
    if IsDirsChild(ListDirName,Self.FileList[I].FilePath) = true then begin
      //.    ...
      if Self.PathsAreAbsolute = true then begin
        //  ...        ...
        AbsoluteName := Self.FileList[I].FilePath;
      end
      else begin
        //  .    ListDirName,    RealDirName
        AbsoluteName := ResolveLocalPath(GetLocalPath(ListDirName, FileList[I].FilePath), RealDirName);
      end;
      //    -   ?
      if Self.FileList[I].FileType = Flist_Filetype_File then begin
        //. ,      .
        if FileExists(AbsoluteName) = true then begin
          //   .   -  ...
          if Self.AddSize = true then begin
            if GetFileSize(AbsoluteName) <> FileList[I].FileSize then begin
              //  . .
              ReserveComparedListMem();
              ComparedList[ListRealSize].FilePath := FileList[I].FilePath;
              ComparedList[ListRealSize].FileType := Flist_Filetype_File;
              ComparedList[ListRealSize].CheckResult := Flist_CheckFile_MismatchedSize;
              ListRealSize := ListRealSize+1;
              MismatchFound := true;
            end;
          end;
          
          //  -  CRC32.
          if (Self.AddCRC32 = true) and (MismatchFound = false) then begin
            if FileToCRC32(AbsoluteName) <> FileList[I].CRC32 then begin
              //  CRC32. .
              ReserveComparedListMem();
              ComparedList[ListRealSize].FilePath := FileList[I].FilePath;
              ComparedList[ListRealSize].FileType := Flist_Filetype_File;
              ComparedList[ListRealSize].CheckResult := Flist_CheckFile_MismatchedCRC32;
              ListRealSize := ListRealSize+1;
              MismatchFound := true;
            end;
          end;
        end
        else begin
          //   ,     .   ,  .
          ReserveComparedListMem();
          ComparedList[ListRealSize].FilePath := FileList[I].FilePath;
          ComparedList[ListRealSize].FileType := Flist_Filetype_File;
          ComparedList[ListRealSize].CheckResult := Flist_CheckFile_FileNotExists;
          ListRealSize := ListRealSize+1;
          MismatchFound := true;
        end;
      end
      else if Self.FileList[I].FileType = Flist_Filetype_Directory then begin
        //. ,      .
        if DirectoryExists(AbsoluteName) = false then begin
          //   ,     .   ,  .
          ReserveComparedListMem();
          ComparedList[ListRealSize].FilePath := FileList[I].FilePath;
          ComparedList[ListRealSize].FileType := Flist_Filetype_Directory;
          ComparedList[ListRealSize].CheckResult := Flist_CheckFile_FileNotExists;
          ListRealSize := ListRealSize+1;
          MismatchFound := true;
        end;
      end;
    end;
  end;
  
  //,  .      ,     , .
  //     ,     .
  RecursiveCompareDir(ListDirName, RealDirName);
  
  //,     .   ""  -.
  SetLength(ComparedList,ListRealSize);
end;

function TFlistFormat.CompareWithFlist(var OtherFlist : TFlistFormat; const ListDirName : AnsiString; var ComparedList : AComparedList) : boolean;
//    ,   (  )        .
var
  ListRealSize, I, SearchElemNum : Integer;
  MismatchFound : Boolean;

  procedure ReserveComparedListMem();
  //  - ""     ComparedList
  begin
    if ListRealSize >= Length(ComparedList) then begin
      //     .
      SetLength(ComparedList, Length(ComparedList)+50);
    end;
  end;

  function ElementReported(const ElemRepName : AnsiString) : Boolean;
  // -         .
  var
    I : Integer;
  begin
    Result := false;
    //...
    for I := 0 to ListRealSize-1 do begin
      if UpperCase(ComparedList[I].FilePath) = UpperCase(ElemRepName) then begin
        //.
        Result := true;
        Exit;
      end;
    end;
  end;

begin
  //.
  Result := false;

  //   ComparedList.
  SetLength(ComparedList,0);

  //   ComparedList.
  ListRealSize := 0;     //      .
  SetLength(ComparedList,Length(Self.FileList)+50);   //    ,         .

  //   ,    "" ...
  for I := 0 to FilesNum-1 do begin
    //  .
    MismatchFound := false;
    // -      ...
    if IsDirsChild(ListDirName,Self.FileList[I].FilePath) = true then begin
      //   ,    ...
      if FileList[I].FileType = Flist_Filetype_Directory then begin
        //. .
        if OtherFlist.DirExistsHere(FileList[I].FilePath) = false then begin
          //    "" .  .
          ReserveComparedListMem();
          ComparedList[ListRealSize].FilePath := FileList[I].FilePath;
          ComparedList[ListRealSize].FileType := Flist_Filetype_Directory;
          ComparedList[ListRealSize].CheckResult := Flist_CheckFile_FileNotExists;
          ListRealSize := ListRealSize+1;
          MismatchFound := true;
        end;
      end
      else begin
        //. .
        SearchElemNum := OtherFlist.GetFileNumberHere(FileList[I].FilePath);
        if SearchElemNum >= 0 then begin
          // .    -  .
          if (Self.AddSize = true) and (OtherFlist.AddSize = true) then begin
            if Self.FileList[I].FileSize <> OtherFlist.FileList[SearchElemNum].FileSize then begin
              //  .  .
              ReserveComparedListMem();
              ComparedList[ListRealSize].FilePath := FileList[I].FilePath;
              ComparedList[ListRealSize].FileType := Flist_Filetype_File;
              ComparedList[ListRealSize].CheckResult := Flist_CheckFile_MismatchedSize;
              ListRealSize := ListRealSize+1;
              MismatchFound := true;
            end;
          end;

          //     -  CRC32.
          if (MismatchFound = false) and (Self.AddCRC32 = true) and (OtherFlist.AddCRC32 = true) then begin
            if Self.FileList[I].CRC32 <> OtherFlist.FileList[SearchElemNum].CRC32 then begin
              //  .  .
              ReserveComparedListMem();
              ComparedList[ListRealSize].FilePath := FileList[I].FilePath;
              ComparedList[ListRealSize].FileType := Flist_Filetype_File;
              ComparedList[ListRealSize].CheckResult := Flist_CheckFile_MismatchedCRC32;
              ListRealSize := ListRealSize+1;
              MismatchFound := true;
            end;
          end;
        end
        else begin
          //  .  .
          ReserveComparedListMem();
          ComparedList[ListRealSize].FilePath := FileList[I].FilePath;
          ComparedList[ListRealSize].FileType := Flist_Filetype_File;
          ComparedList[ListRealSize].CheckResult := Flist_CheckFile_FileNotExists;
          ListRealSize := ListRealSize+1;
          MismatchFound := true;
        end;
      end;
    end;
  end;

  //, .     "" ,      ,  " ".
  for I := 0 to OtherFlist.FilesNum-1 do begin
    // -      ...
    if IsDirsChild(ListDirName,OtherFlist.FileList[I].FilePath) = true then begin
      if ElementReported(OtherFlist.FileList[I].FilePath) = false then begin
        //      ,       ,    ...
        if OtherFlist.FileList[I].FileType = Flist_Filetype_Directory then begin
          //. .
          if Self.DirExistsHere(OtherFlist.FileList[I].FilePath) = false then begin
            //. .
            ReserveComparedListMem();
            ComparedList[ListRealSize].FilePath := OtherFlist.FileList[I].FilePath;
            ComparedList[ListRealSize].FileType := Flist_Filetype_Directory;
            ComparedList[ListRealSize].CheckResult := Flist_CheckFile_ElementNotExists;
            ListRealSize := ListRealSize+1;
          end;
        end
        else begin
          //. .
          if Self.FileExistsHere(OtherFlist.FileList[I].FilePath) = false then begin
            //. .
            ReserveComparedListMem();
            ComparedList[ListRealSize].FilePath := OtherFlist.FileList[I].FilePath;
            ComparedList[ListRealSize].FileType := Flist_Filetype_File;
            ComparedList[ListRealSize].CheckResult := Flist_CheckFile_ElementNotExists;
            ListRealSize := ListRealSize+1;
          end;
        end;
      end;
    end;
  end;

  //,     .   ""  -.
  SetLength(ComparedList,ListRealSize);
  //  .
  Result := true;

end;

function TFlistFormat.CompareWithFlist(const OtherFlistFile, ListDirName : AnsiString; var ComparedList : AComparedList) : boolean;
//    ,   (  )        .
var
  OtherFlist : TFlistFormat;
  
begin
  //.
  Result := false;
  OtherFlist := TFlistFormat.Create;
  
  //.      "" ...
  if OtherFlist.Load(OtherFlistFile) = true then begin
    //        ,   ...
    Result := Self.CompareWithFlist(OtherFlist,ListDirName,ComparedList);
  end;
  
  // ...
  OtherFlist.Free;
end;

function TFlistFormat.ConvertToSimpleComparedList(var ComparedList : AComparedList) : Boolean;
//      -         ,      .
var
  I : Integer;
begin
  //.
  Result := false;
  
  //      ....
  for I := 0 to Length(ComparedList)-1 do begin
    if (ComparedList[I].CheckResult = Flist_CheckFile_MismatchedCRC32) or (ComparedList[I].CheckResult = Flist_CheckFile_MismatchedSize) then begin
      ComparedList[I].CheckResult := Flist_CheckFile_NonIdentical;
    end;
  end;
  
  //.
  Result := true;
end;

////////////////////////////////////////////////////
//          ,   %).     //
////////////////////////////////////////////////////

function Flist_CheckFile_2Text(const Code : Byte) : AnsiString;
//     Flist_CheckFile_*
begin
  Result := 'Unknown check-result code.';
  
  if Code = Flist_CheckFile_Identical then begin
    Result := 'Identical.';
  end
  else if Code = Flist_CheckFile_MismatchedCRC32 then begin
    Result := 'Mismatched CRC32.';
  end
  else if Code = Flist_CheckFile_MismatchedSize then begin
    Result := 'Mismatched size.';
  end
  else if Code = Flist_CheckFile_ElementNotExists then begin
    Result := 'Not exists in Flist.';
  end
  else if Code = Flist_CheckFile_FileNotExists then begin
    Result := 'Disappeared.';
  end
  else if Code = Flist_CheckFile_NothingExists then begin
    Result := 'Not exists.';
  end
  else if Code = Flist_CheckFile_Error then begin
    Result := 'Unknown error.';
  end;
end;

end.

